const express = require('express');
const session = require('express-session');
const axios = require('axios');
const FileStore = require('session-file-store')(session);
const bodyParser = require('body-parser');
const cors = require('cors');
const app = express();
const uuid = require('uuid/v4');
const passport = require('passport');
const LocalStrategy = require('passport-local').Strategy;
const fs = require('fs');
const jsonQuery = require('json-query')
const bcrypt = require('bcrypt-nodejs')
const _ = require('lodash')
const chokidar = require('chokidar');



// Password Hashing using: https://www.browserling.com/tools/bcrypt
// Id generation using: https://www.random.org/strings/?num=10&len=10&digits=on&upperalpha=on&loweralpha=on&unique=on&format=html&rnd=new
// Authentification tutorial: https://medium.com/@evangow/server-authentication-basics-express-sessions-passport-and-curl-359b7456003d

const users=JSON.parse(fs.readFileSync('server/users.json', 'utf8'));
const projects=JSON.parse(fs.readFileSync('server/projects.json', 'utf8'));



// configure passport.js to use the local strategy
passport.use(new LocalStrategy(
  { usernameField: 'email',
    passwordField: 'password' },
  (email, password, done) => {
    var user = jsonQuery(['users[email=?]', email], {data: users}).value //.value[0]
    if(user && email === user.email && bcrypt.compareSync(password, user.password)) {
      return done(null, user, { message: 'Login successful.\n' })
    }
    return done(null, false, { message: 'Invalid username or password.\n' }); //Todo: better output for wrong user/password
  }
));

// tell passport how to serialize the user
passport.serializeUser((user, done) => {
  try {
    done(null, user.id);
  } catch (e) {
    done(e);
  }
});

passport.deserializeUser((id, done) => {
  // const user = users[0].id === id ? users[0] : false; // Todo

  const user = jsonQuery(['users[id=?]', id], {data: users}).value
  done(null, user);
});


app.use(bodyParser.json());
app.use(bodyParser.urlencoded({ extended: false }))

app.use(cors({credentials: true, origin: 'http://localhost:8080'})); //


app.use(session({
  genid: (req) => {
    return uuid() // use UUIDs for session IDs
  },
  store: new FileStore(),
  secret: 'KUW7L4tS9NdE4V20o4VZEw5vswZomIepNbk7AxZKPtF84oUjCnZUpROxJu23',
  resave: true,
  saveUninitialized: true
}))
app.use(passport.initialize());
app.use(passport.session());


// app.all('*', function(req, res, next) {
//   res.header('Access-Control-Allow-Origin', 'http://localhost:8080');
//   res.header('Access-Control-Allow-Credentials', true);
//   res.header('Access-Control-Allow-Methods', 'PUT, GET, POST, DELETE, OPTIONS');
//   res.header('Access-Control-Allow-Headers', 'Content-Type');
//   next();
// });

app.post('/login', (req, res, next) => {
  // console.log(req);

  passport.authenticate('local', (err, user, info) => {
    // if (!user) { res.redirect('/'); }
    req.login(user, (err) => {
      if(err) {
        // console.log(user);
        console.log(err);

        return res.send('Login not successful.\n');
      }
      return res.send('You were authenticated and logged in!\n');
    })

  })(req, res, next);
})

app.get('/userdetails', function(req, res){
  // req.logout();
  // res.redirect('/');
  var user = {
    'isLoggedIn': false,
    'email': 'guest'
  }
  // console.log(req);
  if(req.isAuthenticated()) {
    user = req.user;
    user.isLoggedIn = true;
    // user.password = '';
    user.projects = ['clamy17-dev', 'clamy17', 'demo']; // Todo dynamic
    // console.log(user.groups);
    // projectList = jsonQuery([''], {data: projects}).value; //jsonQuery(['users[id=?]', id], {data: users}).value
    // console.log(projectList);

    user.projects = [];
    projects.forEach( (project) => {
      // console.log('Testing ' + project.id );
      var found = false;
      project.permissions.forEach( (permission) => {
        user.groups.forEach( (group) => {
          // console.log('Found '+ group + ' in ' + permission + ': ' + _.has(permission, group));
          if(_.has(permission, group)) {
            found = true;
          }
        });
        if(found) {
          user.projects.push(project.id )
        }
      });
      user.projects = _.uniq(user.projects).sort();
    });

    console.log(user);

    // user.clientConfig =
    // user.id = '';
    // user.name = req.user.name;
    // console.log('here');
    // console.log(req.user);
  }
  // console.log(user);
  res.send(user);
});

app.get('/logout', function(req, res){
  req.logout();
  req.session.destroy();
  // res.send();
  res.status(401).send('Logged out')
});



// delete below
// app.get('/a', (req, res) => {
//   console.log('Inside GET /authrequired callback')
//   console.log(`User authenticated? ${req.isAuthenticated()}`)
//   if(req.isAuthenticated()) {
//     res.send('you hit the authentication endpoint\n')
//   } else {
//     res.redirect('/')
//   }
// })

app.get('/', (req, res) => {
  res.send('Nothing to see here.')
})

app.use('/api/filesystem', require('./routes/api/filesystem') );
app.use('/api/projects', require('./routes/api/projects') );
app.use('/api/file', require('./routes/api/file') );
app.use('/api/filewatch', require('./routes/api/filewatch') );
app.use('/user', require('./routes/user/user') );



const port = process.env.PORT || 5000;
app.listen(port, () => console.log(`Server started on port ${port}`) );



// Filesystem watching
// https://www.npmjs.com/package/chokidar
// chokidar.watch('.', {ignored: /(^|[\/\\])\../}).on('all', (event, path) => {
//   console.log(event, path);
// });
