const express = require('express');
const router = express.Router();
const fs = require('fs')
const path = require('path')
const _ = require('lodash')

const projectfile=fs.readFileSync('server/projects.json', 'utf8'); // Todo: customization
const projects=JSON.parse(projectfile);


// router.get('/:projectId', async (req,res) => { //async (req,res) => (
//   // res.send('hello!');
//   var fs = await getFilesFromFolder( req.params.projectId ); //await getFilesFromFolder();
//   res.send(await fs);
// });

router.post( '/', async (req,res) => {
  try {
    const result = await getFilesFromFolderWithDetails( req.body.projectId, req.body.wd );
    res.send( result )
  } catch (e) {
    console.log(e);
    res.send( [] ) // Todo send better errors with HTTP codes
  }
});

function getProjectPath(projectId) {
  return _.find(projects, [ 'id', projectId ]).path
}

function getCurrentPath(projectId, wd) {
  const projectPath = getProjectPath(projectId)
  var currentPath = path.resolve( path.join(projectPath, wd) )
  // console.log(currentPath + ' contains ' + projectPath + ' ' + currentPath.startsWith(projectPath))
  if(currentPath.startsWith(projectPath)) {
    return currentPath
  }
  return false;
}

async function getFilesFromFolderWithDetails(projectId, wd) {
  // thanks to https://dustinpfister.github.io/2018/02/15/nodejs-filesystem-stat/
  // const FileHound = require('filehound');

  var result = []
  const currentpath = getCurrentPath(projectId, wd)

  return new Promise(function (resolve, reject) {
    getFilesFromFolder(projectId, wd).then(function (files) {
      var i = 0
      const len = files.length

      next = function () {
        var file = files[i];
        fs.stat(path.join(currentpath, file), function (e, stats) {
          if (e) {
            reject(e.message);
          }
          var fileObj = {
            filename: file,
            isFile: stats.isFile(),
            isDir: stats.isDirectory(),
            isSymbolicLink: stats.isSymbolicLink(),
            isSelected: false,
            validationStatus: 0,
            ext: path.parse(file).ext,
            children: [],
            parents: [],
            stats: stats
          }
          parents = wd.split(path.sep)
          for (var p=0; p<parents.length; p++) {
            var parentObj = {
              filename: parents[p],
              path: path.join( '', _.slice( parents, 0, p+1 ).join(path.sep) )
            }
            fileObj.parents.push(parentObj)
          }

          result.push(fileObj);
          i += 1;
          if (i >= len) {
            resolve(result);
          } else {
            next();
          }

        });

      };
      next();
    }).catch( function(e) {
      reject([]) // Todo: better error handling
    })
  })
}

async function getFilesFromFolder(projectId, wd) {
  const fs = require('fs')
  const path = require('path')
  const FileHound = require('filehound');

  var result = []

  const projectPath = getProjectPath(projectId)
  const currentpath = path.join(projectPath, wd)

  return new Promise(function (resolve, reject) {
    fs.readdir(currentpath, function (error, files) {
      if (error) {
        resolve(result) // Todo: proper error handling
      } else {
        resolve(files)
      }
    })
  });
}

module.exports = router;
