const express = require('express')
const router = express.Router()
const fs = require('fs')
const FileHound = require('filehound')
const path = require('path')
const _ = require('lodash')
const jsonQuery = require('json-query')
const getFolderSize = require('get-folder-size');
const prettyBytes = require('pretty-bytes');

const projectfile=fs.readFileSync('server/projects.json', 'utf8') // Todo: customization
const projects=JSON.parse(projectfile)

const nifti = require('nifti-reader-js')



class sdError extends Error {
  constructor(txt, severity = 0, count = 0, details = '', ...params) {
    super(...params);
    if (Error.captureStackTrace) {
      Error.captureStackTrace(this, sdError);
    }
    this.severity = severity;
    this.txt = txt;
    this.count = count;
    this.date = new Date();
    this.details = details;
  }
}

router.get( '/clear', async (req,res) => {
  try {
    const projectId = req.query.id;
    const resultJSON = "server/projects/" + projectId + ".json";
    if(req.isAuthenticated()) {
      console.log('Deleting file');
      res.send()
      fs.unlinkSync(resultJSON)
    } else {
      res.send()
    }
  } catch (e) {
    res.send( [] ) // Todo send better errors with HTTP codes
  }
})

router.get( '/validate', async (req,res) => {
  // Todo:
  //  Trigger validate on startup and modification
  try {
    const projectId = req.query.id;
    const resultJSON = "server/projects/" + projectId + ".json";
    if(req.isAuthenticated()) {
      if(fs.existsSync(resultJSON)) {
        const result=fs.readFileSync(resultJSON, 'utf8');
        res.send(result);
      } else {
        console.log('Starting new validation');
        var result = await create(projectId);
        result = await validate(result, projectId);
        fs.writeFile(resultJSON, JSON.stringify(result), function(err) {
          if (err) throw err;
          }
        );
        console.log('Finished new validation');
        res.send(result)
      }
    } else {
      res.send()
    }

  } catch (e) {
    res.send( [] ) // Todo send better errors with HTTP codes
  }
})

// router.get( '/validateDev', async (req,res) => {
//   // Todo:
//   //  Trigger validate on startup and modification
//   try {
//     const projectId = req.query.id;
//     var result = await create(projectId)
//     result = await validate(result, projectId)
//     res.send( '<pre>'+JSON.stringify(result, null, 2)+'</pre>' )
//   } catch (e) {
//     console.log(e);
//     res.send( [] ) // Todo send better errors with HTTP codes
//   }
// })

async function create(projectId, parent=false, relPath=false, objPath='', projectJSON=false) {
  return new Promise( async (resolve, reject) => {
    var res = {}
    try {
      if(!relPath) {
        relPath = '/'
      }
      absPath = path.join(getProjectPath(projectId),relPath)
      if(!projectJSON) {
        projectJSON = JSON.parse( fs.readFileSync(path.join(getProjectPath(projectId), projectId+'.sweetData.json' ), 'utf8' ) )
      }

      res = {
        fullpath:   relPath,
        objpath:    objPath+'/',
        filename:   ( path.parse(relPath).ext ? path.parse(relPath).name + path.parse(relPath).ext : path.parse(relPath).name ),
        path:       path.parse(relPath).dir,
        ext:        path.parse(relPath).ext,
        size:       false,
        sizeRequired: 0,
        sizeCorrect:  true,
        dateCreated:  0,
        dateModified: 0,
        dateAccessed: 0,
        parent:     parent,
        sdEntity:   false,
        sdWatch:    false,
        sdTidy:     false,
        sdChildren: [],
        sdMissingChildren: [],
        sdUnknownChildren: [],
        sdValidationErrors: [],
        sdValidationErrorsCount: [0, 0, 0, 0],
        sdRequired: true,
        sdIsMissing: false,
        sdTypeInfo: {}
      };

      stats = fs.statSync(absPath)
      res.dateCreated = stats.birthtimeMs;
      if(stats.isDirectory()) {
        res.type = 'folder'
      } else {
        res.type = 'file';
        projectJSON.types.forEach( (type) => {
          if(type.ext.indexOf(_.trimStart(res.ext,'.')) >= 0) {
            // console.log('Found ' + res.ext + ' is ' + type.name);
            res.type = type.name;
          }
        })
        res.size = stats.size;
        res.dateModified = stats.mtimeMs;
        res.dateAccessed = stats.atimeMs;
      }

      [ res.sdEntity, res.sdWatch ] = await identify(projectJSON, res);
      res.objpath += res.sdEntity;
      if(res.sdEntity && res.sdEntity[0] != '@') {
        res.sizeRequired = jsonQuery(['[**][name=? ]', res.sdEntity], {data: projectJSON}).value.filesize_mb*1024*1024;
        res.sizeCorrect = ( res.sizeRequired < res.size ? true : false );
        res.sdTidy = jsonQuery(['[**][name=? ]', res.sdEntity], {data: projectJSON}).value.tidy;
      }


      if(res.type==='folder')
        if(res.sdWatch) {
          var children = fs.readdirSync(absPath);
          children = children.filter(item => !(/(^|\/)\.[^\/\.]/g).test(item));
          children = children.filter(item => !(/.lnk$/g).test(item));

          var sdChildren = []
          for(var i=0; i<children.length; i++ ) {
            var childPath = path.join(res.fullpath,children[i]);
            sdChildren[i] = await create(projectId, res.sdEntity, childPath, res.objpath, projectJSON)
          }
          res.size = sdChildren.map( child => child.size ).reduce((a, b) => a + b, 0);
          res.dateAccessed = Math.max(...sdChildren.map( child => child.dateAccessed ));
          res.dateModified = Math.max(...sdChildren.map( child => child.dateModified ));

          [ res.sdChildren, res.sdMissingChildren, res.sdUnknownChildren ] = sortChildren( res.sdEntity, sdChildren, JSON.parse(JSON.stringify(projectJSON)) ); // deep copy needed
        } else {
          let promise = new Promise((resolve, reject) => {
            getFolderSize(absPath, async (err, size) => {
              if (err) { throw err; }
              resolve(size);
            });
          });

          res.size = await promise;
        }

      switch(res.type) {
        case 'nifti':
        fs.readFile(absPath, (err, buf) => {
          if(typeof(buf)==='undefined') {
            res.sdValidationErrors.push(
              new sdError( 'File not accessible: '+  filepath , 3, 1 )
            )
            res.sdValidationErrorsCount[3]+=1
          }
          var data = nifti.Utils.toArrayBuffer(buf)
          var niftiHeader = null,
          niftiImage = null,
          niftiExt = null;
          if(nifti.isCompressed(data)) {
            data = nifti.decompress(data);
          }
          if(nifti.isNIFTI(data)) {
            niftiHeader = nifti.readHeader(data);
            res.sdTypeInfo.niiVolumes = niftiHeader.dims[4];
          }
          if(res.sdEntity[0]!=='@') {
            res.sdTypeInfo.niiVolumesRequired = jsonQuery(['[**][name=? ]', res.sdEntity], {data: projectJSON}).value.nii_volumes;
          }
        });
        break;
      }
      resolve(res)
    } catch(e) {
      reject(e);
    }
  })
}

async function validate(req, projectId) {
  // Todo:
  //   Externalize validation of filetype specific features

  return new Promise( async (resolve, reject) => {
    var res = req;
    if(res.type === 'folder' ) {
      // ERROR    folders have missing
      if(res.sdMissingChildren && Object.keys(res.sdMissingChildren).length>0) {
        Object.keys(res.sdMissingChildren).forEach( (key) => {
          if(res.sdMissingChildren[key][0]>0) {
            res.sdValidationErrors.push(
              new sdError('Required objects ' + key + ' not found. ' +  res.sdMissingChildren[key][0] + ' more missing ('+ res.sdMissingChildren[key][1] +' more expected).', 3, res.sdMissingChildren[key][0])
            )
            res.sdValidationErrorsCount[3]+=res.sdMissingChildren[key][0]
          } else if(res.sdMissingChildren[key][1]>0) {
            res.sdValidationErrors.push(
              new sdError('Optional objects ' + key + ' not found. ' +  res.sdMissingChildren[key][1] + ' more expected.', 2, res.sdMissingChildren[key][1])
            )
            res.sdValidationErrorsCount[2]++
          }
        })
      }

      // WARNING  folders have unknown children
      var unknownObjects=_.filter( res.sdChildren, ['sdEntity[0]', '@'] );
      if(unknownObjects.length>0) {
        if(res.sdTidy) {
          var e = new sdError( unknownObjects.length + ' unknown objects found. Folder should be tidy.', 2, unknownObjects.length, unknownObjects );
        } else {
          var e = new sdError( unknownObjects.length + ' unknown objects found.', 1, unknownObjects.length, unknownObjects );
        }
        res.sdValidationErrors.push(e)
        res.sdValidationErrorsCount[1]+=unknownObjects.length
      }

      for(var i=0; i<res.sdChildren.length; i++) {
        res.sdChildren[i] = await validate( res.sdChildren[i], projectId )
      }
    }

    // ERROR    filesize_mb too small
    if(res.size<res.sizeRequired) {
      res.sdValidationErrors.push(
        new sdError( 'Size too small. Expected '+  prettyBytes(res.sizeRequired) +' but found '+ prettyBytes(res.size) +' ('+ prettyBytes(res.sizeRequired-res.size) +' missing).', 3, 1 )
      )
      res.sdValidationErrorsCount[3]+=1
    }

    if(res.type==='nifti') {
      if(res.sdTypeInfo.niiVolumesRequired) {
        if(res.sdTypeInfo.niiVolumesRequired>res.sdTypeInfo.niiVolumes) {
          res.sdValidationErrors.push(
            new sdError( 'Volume too small. Expected '+ res.sdTypeInfo.niiVolumesRequired +' but found '+ res.sdTypeInfo.niiVolumes +' volumes.', 3, 1 )
          )
          res.sdValidationErrorsCount[3]+=1
        }
      }

    }

    for(var i=0; i<res.sdChildren.length; i++) {
      for( var ei=0; ei<res.sdChildren[i].sdValidationErrorsCount.length; ei++) {
        res.sdValidationErrorsCount[ei] += res.sdChildren[i].sdValidationErrorsCount[ei];
      }
    }
    resolve(res)
  })

}

async function identify(projectJSON, res) {
  return new Promise( async (resolve, reject) => {

    var candidates = projectJSON.entities;
    candidates = jsonQuery(['[*][* parent=? | parent=*]', res.parent], {data: candidates}).value    // Test: parent directory
    candidates = jsonQuery(['[*][* type=? | type=*]', res.type], {data: candidates}).value          // Test: filetype

    var newCandidates = [];
    candidates.forEach( (candidate, i, obj) => {
      var patternString = candidate.pattern
      var varPattern = new RegExp('\\${.*}')
      if(varPattern.test( candidate.pattern )) {
        varPattern.exec( candidate.pattern ).forEach( (v) => {
          var replaceV = res.fullpath.split(path.sep)[ res.objpath.split('/').indexOf(_.trimEnd(_.trimStart(v,'${'),'}'))-1 ];
           patternString = candidate.pattern.replace(v, replaceV);
        });
      }

      pattern = new RegExp(patternString)
      if( res.filename === patternString |  pattern.test(res.filename) ) {
        newCandidates.push( candidate )
      }
    })
    candidates = newCandidates

    if(candidates.length == 1) {
      resolve([candidates[0].name, candidates[0].watch])
    }
    if(candidates.length > 1) {
      resolve(['@ambiguous: ' + _.map(candidates, 'name'), false ])
    }
    resolve([ '@unknown', false ])
  })
}

function sortChildren(sdEntity, sdChildren, projectJSON) {
  if(sdEntity && sdEntity[0] != '@' ) {
    var sdMissingChildren = jsonQuery(['[**][* name=? ]', sdEntity], {data: projectJSON}).value[0].contains
        sdUnknownChildren = []
        children          = sdChildren
        sdChildren        = [];

    if(sdMissingChildren) {
      Object.keys(sdMissingChildren).forEach( (child, i, obj) => {
        var missingchild = jsonQuery(['[**][* name=? ]', child], {data: projectJSON}).value[0]
        var res = {
          fullpath:   '',
          objpath:    '',
          filename:   missingchild.pattern,
          ext:        '',
          size:       false,
          sizeRequired: 0,
          sizeCorrect:  true,
          dateCreated:  0,
          dateModified: 0,
          dateAccessed: 0,
          parent:     missingchild.parent,
          sdEntity:   child,
          sdWatch:    false,
          sdChildren: [],
          sdMissingChildren: [],
          sdUnknownChildren: [],
          sdValidationErrors: [],
          sdValidationErrorsCount: [0, 0, 0, 0],
          sdRequired: true,
          sdIsMissing: true
        }

        if(!Array.isArray(sdMissingChildren[child])) {
          sdMissingChildren[child] = [sdMissingChildren[child], sdMissingChildren[child]]
        }
        for(var i=0; i<sdMissingChildren[child][0]; i++) {
          sdChildren.push(JSON.parse(JSON.stringify(res)));
        }
        for(var i=0; i<sdMissingChildren[child][1]-sdMissingChildren[child][0]; i++) {
          res.sdRequired = false;
          sdChildren.push(JSON.parse(JSON.stringify(res)));
        }

      })
      }

      children.forEach( (child, i, obj) => {
        var index=_.findIndex(sdChildren, function(o) { return o.sdEntity == child.sdEntity && o.sdIsMissing; }); //
        if(index>=0 ) {
          sdChildren[index] = child;
          sdMissingChildren[child.sdEntity][0]--;
          sdMissingChildren[child.sdEntity][1]--;
        } else {
          sdChildren.push(child);
        }
      })

    // keep for reference
    // sdChildren = _.sortBy(sdChildren, ['sdEntity', 'filename']);
    // sdUnknownChildren = _.sortBy(sdUnknownChildren, ['filename']);
    // sdChildren = sdChildren.concat(sdUnknownChildren);
    // sdUnknownChildren = [];

    sdUnknownChildren = _.sortBy(_.filter(sdChildren, function(o) { return o.sdEntity[0] === '@'; }), ['sdEntity']);
    sdChildren = _.sortBy(_.filter(sdChildren, function(o) { return o.sdEntity[0] !== '@'; }), ['sdEntity', 'filename']);
    sdChildren = sdChildren.concat(sdUnknownChildren);
    sdUnknownChildren = [];

    return [sdChildren, sdMissingChildren, sdUnknownChildren]
  }
  return [[], [], sdChildren]
}

function getProjectPath(projectId) {
  return _.find(projects, [ 'id', projectId ]).path
}



module.exports = router;




























//
// async function getFilesFromFolder(projectId, wd) {
//   const fs = require('fs')
//   const path = require('path')
//
//
//   var result = []
//
//   const projectPath = getProjectPath(projectId)
//   const currentpath = path.join(projectPath, wd)
//
//   // console.log( 'Listing files for ' + projectId + ' ' + wd )
//
//   return new Promise(function (resolve, reject) {
//     fs.readdir(currentpath, function (error, files) {
//       if (error) {
//         resolve(result) // Todo: proper error handling
//       } else {
//         resolve(files)
//       }
//     })
//   });
// }
//
// function getProjectPath(projectId) {
//   // try {
//     return _.find(projects, [ 'id', projectId ]).path
//   // }
//
//   // catch (e) {
//   //   return false;
//   // }
// }
//
// function getCurrentPath(projectId, wd) {
//   const projectPath = getProjectPath(projectId)
//   var currentPath = path.resolve( path.join(projectPath, wd) )
//   // console.log(currentPath + ' contains ' + projectPath + ' ' + currentPath.startsWith(projectPath))
//   if(currentPath.startsWith(projectPath)) {
//     return currentPath
//   } else {
//     throw "Requested path outside project directory."
//   }
//
// }
//
// async function getFolders(projectId, wd, withDetails = true) {
//   var filesystem = []
//   const currentpath = getCurrentPath(projectId, wd)
//   const projectPath = getProjectPath(projectId)
//
//   const FileHound = require('filehound');
//   // const projectPath = getProjectPath(projectId)
//   const folders = FileHound.create()
//   .paths(currentpath)
//   .depth(1)
//   // .directory()
//   .find();
//
//   return new Promise(function (resolve, reject) {
//     folders.then( async function (folders) {
//       var i = 0
//
//       next = async function () {
//         var folder = folders[i]
//         var details = {}
//         details = {
//           filename:   path.parse(folder).name,
//           fullpath:   folder.replace(projectPath,''),
//           path:       fullpath.replace(filename,''),
//           ext:        path.parse(folder).ext,
//           isFile:     false,
//           isDir:      true,
//           parentpath: false,
//         }
//         filesystem.push( details )
//         i += 1;
//         if (i >= folders.length) {
//           resolve(filesystem)
//         } else {
//           next();
//         }
//       }
//       next();
//     })
//   })
// }
//
// // console.log('Get Folders: ' + projectId +'  '+ wd + '  in  ' + currentpath)
//
//
// // console.log(folders);
//
// // return new Promise(function (resolve, reject) {
// //   folders.then( async function (files) {
// //     // console.log(files);
// //     console.log(files)
// //     var i = 0
// //     const len = files.length
// //     next = async function () {
// //       var file = files[i];
// //       try {
// //         var details = '';
// //         console.log('Getting details for ' + file);
// //         if(typeof(file)=='string') {
// //           if(withDetails) {
// //             details = {} //await getFolderDetails(file, projectId, wd)
// //           } else {
// //             details = {'fullpath': file}
// //           }
// //           result.push( details )
// //         }
// //       } catch(e) {
// //         console.log(e);
// //       }
// //
// //
// //       i += 1;
// //       if (i >= len) {
// //         // console.log(result);
// //         resolve(result);
// //       } else {
// //         next();
// //       }
// //     }
// //     next();
// //   })
// // })
// // }
//
// async function getFolderDetails(filepath, projectId, wd) {
//
//
//   const projectPath = getProjectPath(projectId)
//   return new Promise(async function (resolve, reject) {
//
//     if(typeof(filepath)!='string') {
//       reject([]);
//     } else {
//       var result = {
//         filename:   path.parse(filepath).name,
//         fullpath:   filepath.replace(projectPath,''),
//         path:       fullpath.replace(filename,''),
//         ext:        false,
//         isFile:     false,
//         isDir:      true,
//         parentpath: false,
//         parents:    [],
//         children:   [],
//       }
//       result.parentpath = result.fullpath.replace(result.filename,'')
//
//       parents = result.parentpath.split(path.sep)
//       for (var p=0; p<parents.length-1; p++) {
//         var parentObj = {
//           filename: parents[p],
//           fullpath: path.join( path.sep, _.slice( parents, 0, p+1 ).join(path.sep) )
//         }
//         result.parents.push(parentObj)
//       }
//       // try {
//       //   console.log('Looking for children ' + projectId +'  '+ result.fullpath )
//       //   childrenPromise = getFolders(projectId, result.fullpath, true)
//       //   childrenPromise.then( function (children) {
//       //     console.log(children);
//       //     console.log('Done looking for children ' + projectId +'  '+ result.fullpath );
//       //     result.children = children
//       //   })
//       //   // next = async function () {
//       //   // }
//       //
//       //
//       //   // children.then(
//       //   //   async function (children) {
//       //   //     console.log(children);
//       //   //   }
//       //   // )
//       // } catch(e) {
//       //   console.log('No children');
//       //   result.children = []
//       // }
//       //
//
//       resolve(result);
//
//     }
//
//   })
// }
//
// async function getChildren (projectId, result) {
//   return new Promise(async function (resolve, reject) {
//     // for(var i=0; p<result.length; i++) {
//     //   result[i].children = [ await getFolders(projectId, result[i].fullpath, false) ]
//     // }
//     // console.log(result[r].children[i]);
//     for(var r=0; r<result.length; r++) {
//       var i = 0
//       var children = await getFolders(projectId, result[r].fullpath, false)
//       const len = children.length; //result[i].children.length
//       next = async function () {
//         result[r].children[i] = children[i]
//         try {
//           var i2 = 0
//           var children2 = await getChildren(projectId, result[r].children[i].fullpath)
//           const len = children.length; //result[i].children.length
//           next2 = async function () {
//             result[r].children[i].children[i2] = children2
//             i2 += 1;
//             if (i2 < len) {
//               next2();
//             }
//           }
//           next2();
//           // result[r].children[i].children = 'foo' //await getChildren(projectId, result[r].children[i].fullpath)
//         } catch(e) {
//           console.log(e);
//         }
//
//         i += 1;
//         if (i < len) {
//           next();
//         }
//       }
//       next();
//       // for(var c=0; c<children.length; c++) {
//       //   result[r].children[i] = children[i]
//       // }
//
//
//       // next = async function () {
//       //     // var file = files[i];
//       //     // try {
//       //     //   var details = '';
//       //     //   console.log('Getting details for ' + file);
//       //     //   if(typeof(file)=='string') {
//       //     //     if(withDetails) {
//       //     //       details = await getFolderDetails(file, projectId, wd)
//       //     //     } else {
//       //     //       details = {'file': file}
//       //     //     }
//       //     //     result.push( details )
//       //     //   }
//       //     // } catch(e) {
//       //     //   console.log(e);
//       //     // }
//       //
//       //     result[r].children[i] = children[i]
//       //     i += 1;
//       //     if (i >= len) {
//       //       // resolve(result);
//       //     } else {
//       //       next();
//       //     }
//       // }
//       // next();
//
//     }
//     resolve( result )
//   })
// }
//
//
//
//
//
//
//
//
//
//
//
//
//
//
//
//
//
//
//
//
//
//
//
//
//
//
//
//
//
//
//
//
//
//
//
//
//
//
//
//
//
// async function getFilesFromFolderWithDetails3(projectId, wd) {
//   const FileHound = require('filehound');
//   const projectPath = getProjectPath(projectId)
//   const files = FileHound.create()
//   .paths(path.join(projectPath, wd))
//   .find();
//
//   var result = []
//   files.then( function (files) {
//     // console.log(files);
//     i=0;
//     var file = files[i];
//     // crntP = path.join(currentpath, file)
//     fs.stat(file, function (e, stats) {
//       if (e) {
//         reject(e.message);
//       }
//       var fileObj = {
//         filename: file,
//         fullpath: file,
//         isFile: stats.isFile(),
//         isDir: stats.isDirectory(),
//         isSymbolicLink: stats.isSymbolicLink(),
//         isSelected: false,
//         validationStatus: 0,
//         ext: path.parse(file).ext,
//         children: [],
//         parentpath: '',
//         parents: [],
//         //stats: stats
//       }
//       result.push(fileObj);
//     })
//   })
// }
//
//
//
//
// async function getFilesFromFolderWithDetails(projectId, wd) {
//   // thanks to https://dustinpfister.github.io/2018/02/15/nodejs-filesystem-stat/
//   // const FileHound = require('filehound');
//
//   var result = []
//   const currentpath = getCurrentPath(projectId, wd)
//
//   const FileHound = require('filehound');
//   const projectPath = getProjectPath(projectId)
//   const files = FileHound.create()
//   .paths(path.join(projectPath, wd))
//   .depth(0)
//   .directory()
//   .find();
//
//   return new Promise(function (resolve, reject) {
//
//     files.then( function (files) {
//
//       // getFilesFromFolder(projectId, wd).then(function (files) {
//       // console.log(files.length);
//       // if(files.length > 0) {
//       var i = 0
//       const len = files.length
//       // if(len==0) {
//       //   console.log('0 files found.');
//       //   // resolve(result);
//       //   return;
//       // }
//       // console.log(len);
//       next = function () {
//         var file = files[i];
//         // if(typeof(file)==='string') {
//         // crntP = file; //path.join(currentpath, file)
//         // console.log(file);
//         fs.stat(file, function (e, stats) {
//           if (e) {
//             reject(e.message);
//           }
//           var fileObj = {
//             filename: file,
//             fullpath: file,
//             isFile: stats.isFile(),
//             isDir: stats.isDirectory(),
//             isSymbolicLink: stats.isSymbolicLink(),
//             isSelected: false,
//             validationStatus: 0,
//             ext: path.parse(file).ext,
//             children: [],
//             parentpath: '',
//             parents: [],
//             //stats: stats
//           }
//           parents = wd.split(path.sep)
//           for (var p=0; p<parents.length; p++) {
//             var parentObj = {
//               filename: parents[p],
//               path: path.join( '', _.slice( parents, 0, p+1 ).join(path.sep) )
//             }
//             fileObj.parents.push(parentObj)
//           }
//           fileObj.parentpath = _.slice( fileObj.parents, -1 )[0].path
//           fileObj.fullpath   = path.join(fileObj.parentpath, file)
//
//           // if(fileObj.isDir) {
//           //   try {
//           //     fileObj.children = getFilesFromFolderWithDetails(projectId, fileObj.fullpath);
//           //   } catch (e) {
//           //     console.log(e);
//           //     fileObj.children = [] // Todo better error handling
//           //   }
//           // }
//           result.push(fileObj);
//
//           i += 1;
//           if (i >= len) {
//             resolve(result);
//           } else {
//             next();
//           }
//
//         });
//         // }
//       };
//       next();
//       // }
//     }).catch( function(e) {
//       reject([]) // Todo: better error handling
//     })
//   })
// }
